@@ -17,6 +17,7 @@
 #include <ucs/sys/compiler.h>
 #include <ucs/sys/string.h>
 #include <ucs/vfs/base/vfs_obj.h>
+#include <ucs/vfs/sock/vfs_sock.h>
 #include <sys/signal.h>
 
 
@@ -74,6 +75,22 @@ static UCS_CONFIG_DEFINE_ARRAY(signo,
                                UCS_CONFIG_TYPE_SIGNO);
 
 
+#define UCS_DISTANCE_KEYS_DESCRIPTION(_field) \
+    {"phb", \
+     "connection traversing PCIe as well as a PCIe Host Bridge (typically " \
+     "the CPU)", \
+     ucs_offsetof(ucs_global_opts_t, dist.phb._field)}, \
+    {"node", \
+     "connection traversing PCIe as well as the interconnect between PCIe " \
+     "Host Bridges within a NUMA node", \
+     ucs_offsetof(ucs_global_opts_t, dist.node._field)}, \
+    {"sys", \
+     "connection traversing PCIe as well as the SMP interconnect between " \
+     "NUMA nodes", \
+     ucs_offsetof(ucs_global_opts_t, dist.sys._field)}, \
+    {NULL}
+
+
 static ucs_config_field_t ucs_global_opts_table[] = {
  {"LOG_LEVEL", "warn",
   "UCS logging level. Messages with a level higher or equal to the selected "
@@ -191,6 +208,16 @@ static ucs_config_field_t ucs_global_opts_table[] = {
   "The list order decides the priority of the providers.",
   ucs_offsetof(ucs_global_opts_t, topo_prio), UCS_CONFIG_TYPE_STRING_ARRAY},
 
+ {"DISTANCE_LAT", "phb:300ns,node:300ns,sys:500ns",
+  "Estimated latency between system devices", 0,
+  UCS_CONFIG_TYPE_KEY_VALUE(UCS_CONFIG_TYPE_TIME,
+                            UCS_DISTANCE_KEYS_DESCRIPTION(latency))},
+
+ {"DISTANCE_BW", "auto",
+  "Estimated bandwidth between system devices", 0,
+  UCS_CONFIG_TYPE_KEY_VALUE(UCS_CONFIG_TYPE_BW,
+                            UCS_DISTANCE_KEYS_DESCRIPTION(bandwidth))},
+
  {NULL}
 };
 
@@ -225,6 +252,10 @@ static ucs_config_field_t ucs_global_opts_read_only_table[] = {
   "Enable virtual monitoring filesystem",
   ucs_offsetof(ucs_global_opts_t, vfs_enable), UCS_CONFIG_TYPE_BOOL},
 
+  {"VFS_SOCK_PATH", UCX_VFS_SOCK_DEFAULT_PATH,
+   "Listening UNIX socket path of the VFS daemon.",
+   ucs_offsetof(ucs_global_opts_t, vfs_sock_path), UCS_CONFIG_TYPE_STRING},
+
  {"VFS_THREAD_AFFINITY", "n",
   "Enable inheriting main process affinity for virtual monitoring filesystem\n"
   "service thread. Setting this value to 'n' will allow the service thread to\n"
