@@ -80,7 +80,7 @@ ucs_status_t ucs_netif_ioctl(const char *if_name, unsigned long request,
 
     ucs_strncpy_zero(if_req->ifr_name, if_name, sizeof(if_req->ifr_name));
 
-    status = ucs_socket_create(AF_INET, SOCK_STREAM, &fd);
+    status = ucs_socket_create(AF_INET, SOCK_STREAM, 0, &fd);
     if (status != UCS_OK) {
         goto out;
     }
@@ -195,9 +195,9 @@ unsigned ucs_netif_bond_ad_num_ports(const char *bond_name)
     return (unsigned)ad_num_ports;
 }
 
-ucs_status_t ucs_socket_create(int domain, int type, int *fd_p)
+ucs_status_t ucs_socket_create(int domain, int type, int protocol, int *fd_p)
 {
-    int fd = socket(domain, type, 0);
+    int fd = socket(domain, type, protocol);
     if (fd < 0) {
         ucs_socket_print_error_info("socket create failed", errno);
         return UCS_ERR_IO_ERROR;
@@ -451,7 +451,7 @@ ucs_status_t ucs_socket_server_init(const struct sockaddr *saddr, socklen_t sock
 
     /* Create the server socket for accepting incoming connections */
     fd     = -1; /* Suppress compiler warning */
-    status = ucs_socket_create(saddr->sa_family, SOCK_STREAM, &fd);
+    status = ucs_socket_create(saddr->sa_family, SOCK_STREAM, 0, &fd);
     if (status != UCS_OK) {
         goto err;
     }
@@ -542,7 +542,7 @@ ucs_socket_handle_io_error(int fd, const char *name, ssize_t io_retval, int io_e
  * @param [in]  fd         The socket fd.
  * @param [in]  data       The pointer to user's data or pointer to the array of
  *                         iov elements.
- * @param [in]  count      The length of user's data or the number of elemnts in
+ * @param [in]  count      The length of user's data or the number of elements in
  *                         the array of iov.
  * @param [out] length_p   Pointer to the result length of user's data that was
  *                         sent/received.
@@ -582,9 +582,9 @@ ucs_socket_handle_io(int fd, const void *data, size_t count,
 
 static inline ucs_status_t
 ucs_socket_do_io_nb(int fd, void *data, size_t *length_p,
-                    ucs_socket_io_func_t io_func, const char *name)
+                    ucs_socket_io_func_t io_func, const char *name, int flags)
 {
-    ssize_t ret = io_func(fd, data, *length_p, MSG_NOSIGNAL);
+    ssize_t ret = io_func(fd, data, *length_p, MSG_NOSIGNAL | flags);
     return ucs_socket_handle_io(fd, data, *length_p, length_p, 0,
                                 ret, errno, name);
 }
@@ -597,7 +597,7 @@ ucs_socket_do_io_b(int fd, void *data, size_t length,
     ucs_status_t status;
 
     do {
-        status = ucs_socket_do_io_nb(fd, data, &cur_cnt, io_func, name);
+        status = ucs_socket_do_io_nb(fd, data, &cur_cnt, io_func, name, 0);
         done_cnt += cur_cnt;
         ucs_assert(done_cnt <= length);
         cur_cnt = length - done_cnt;
@@ -624,7 +624,7 @@ ucs_socket_do_iov_nb(int fd, struct iovec *iov, size_t iov_cnt, size_t *length_p
 ucs_status_t ucs_socket_send_nb(int fd, const void *data, size_t *length_p)
 {
     return ucs_socket_do_io_nb(fd, (void*)data, length_p,
-                               (ucs_socket_io_func_t)send, "send");
+                               (ucs_socket_io_func_t)send, "send", 0);
 }
 
 /* recv is declared as 'always_inline' on some platforms, it leads to
@@ -634,9 +634,10 @@ static ssize_t ucs_socket_recv_io(int fd, void *data, size_t size, int flags)
     return recv(fd, data, size, flags);
 }
 
-ucs_status_t ucs_socket_recv_nb(int fd, void *data, size_t *length_p)
+ucs_status_t ucs_socket_recv_nb(int fd, void *data, int flags, size_t *length_p)
 {
-    return ucs_socket_do_io_nb(fd, data, length_p, ucs_socket_recv_io, "recv");
+    return ucs_socket_do_io_nb(fd, data, length_p, ucs_socket_recv_io,
+                               "recv", flags);
 }
 
 ucs_status_t ucs_socket_send(int fd, const void *data, size_t length)
@@ -951,6 +952,38 @@ int ucs_sockaddr_ip_cmp(const struct sockaddr *sa1, const struct sockaddr *sa2)
                   UCS_IPV4_ADDR_LEN : UCS_IPV6_ADDR_LEN);
 }
 
+int ucs_sockaddr_is_same_subnet(const struct sockaddr *sa1,
+                                const struct sockaddr *sa2, unsigned prefix_len)
+{
+    const void *ipaddr1, *ipaddr2;
+    size_t addr_size, addr_size_bits;
+
+    if (sa1->sa_family != sa2->sa_family) {
+        ucs_debug("different addr_family: s1 %s s2 %s",
+                  ucs_sockaddr_address_family_str(sa1->sa_family),
+                  ucs_sockaddr_address_family_str(sa2->sa_family));
+        return 0;
+    }
+
+    /* Get address size */
+    if (ucs_sockaddr_inet_addr_sizeof(sa1, &addr_size) != UCS_OK) {
+        return 0;
+    }
+
+    addr_size_bits = addr_size * CHAR_BIT;
+
+    /* Truncate prefix_len if it exceeds address size */
+    prefix_len = ucs_min(prefix_len, addr_size_bits);
+
+    ipaddr1 = ucs_sockaddr_get_inet_addr(sa1);
+    ipaddr2 = ucs_sockaddr_get_inet_addr(sa2);
+    ucs_assertv((ipaddr1 != NULL) && (ipaddr2 != NULL), "ipaddr1=%p ipaddr2=%p",
+                ipaddr1, ipaddr2);
+
+    /* Check if the addresses have matching prefixes */
+    return ucs_bitwise_is_equal(ipaddr1, ipaddr2, prefix_len);
+}
+
 ucs_status_t ucs_sockaddr_set_inaddr_any(struct sockaddr *saddr, sa_family_t af)
 {
     struct sockaddr_in *sa_in;
