@@ -48,6 +48,40 @@ typedef struct uct_cuda_ipc_remote_cache {
 
 uct_cuda_ipc_remote_cache_t uct_cuda_ipc_remote_cache;
 
+#if HAVE_CUDA_FABRIC
+static UCS_F_ALWAYS_INLINE int
+uct_cuda_ipc_rem_mpool_cache_hash_equal(CUmemFabricHandle key1,
+                                        CUmemFabricHandle key2)
+{
+    return !memcmp(&key1, &key2, sizeof(key1));
+}
+
+/* djb2 hash function */
+static UCS_F_ALWAYS_INLINE khint_t
+uct_cuda_ipc_rem_mpool_cache_hash_func(CUmemFabricHandle key)
+{
+    khint_t hash = 5381;
+    int i;
+
+    for (i = 0; i < sizeof(key.data); i++) {
+        hash = ((hash << 5) + hash) + key.data[i];
+    }
+
+    return hash;
+}
+
+KHASH_INIT(cuda_ipc_rem_mpool_cache, CUmemFabricHandle, CUmemoryPool, 1,
+           uct_cuda_ipc_rem_mpool_cache_hash_func,
+           uct_cuda_ipc_rem_mpool_cache_hash_equal)
+
+typedef struct {
+    khash_t(cuda_ipc_rem_mpool_cache) hash;
+    pthread_rwlock_t                  lock;
+} uct_cuda_ipc_rem_mpool_cache_t;
+
+static uct_cuda_ipc_rem_mpool_cache_t uct_cuda_ipc_rem_mpool_cache;
+#endif
+
 static ucs_pgt_dir_t *uct_cuda_ipc_cache_pgt_dir_alloc(const ucs_pgtable_t *pgtable)
 {
     void *ptr;
@@ -77,6 +111,31 @@ uct_cuda_ipc_cache_region_collect_callback(const ucs_pgtable_t *pgtable,
     ucs_list_add_tail(list, &region->list);
 }
 
+static ucs_status_t uct_cuda_ipc_close_memhandle(uct_cuda_ipc_cache_region_t *region)
+{
+    ucs_status_t status;
+
+#if HAVE_CUDA_FABRIC
+    if (region->key.ph.handle_type == UCT_CUDA_IPC_KEY_HANDLE_TYPE_VMM) {
+        status = UCT_CUDADRV_FUNC_LOG_WARN(cuMemUnmap(
+                    (CUdeviceptr)region->mapped_addr, region->key.b_len));
+        if (status == UCS_OK) {
+            return UCT_CUDADRV_FUNC_LOG_WARN(cuMemAddressFree(
+                        (CUdeviceptr)region->mapped_addr, region->key.b_len));
+        }
+    } else if (region->key.ph.handle_type == UCT_CUDA_IPC_KEY_HANDLE_TYPE_MEMPOOL) {
+        return UCT_CUDADRV_FUNC_LOG_WARN(
+                cuMemFree((CUdeviceptr)region->mapped_addr));
+    } else
+#endif
+    {
+        return UCT_CUDADRV_FUNC_LOG_WARN(cuIpcCloseMemHandle(
+                    (CUdeviceptr)region->mapped_addr));
+    }
+
+    return status;
+}
+
 static void uct_cuda_ipc_cache_purge(uct_cuda_ipc_cache_t *cache)
 {
     int active = uct_cuda_base_is_context_active();
@@ -88,40 +147,220 @@ static void uct_cuda_ipc_cache_purge(uct_cuda_ipc_cache_t *cache)
                       &region_list);
     ucs_list_for_each_safe(region, tmp, &region_list, list) {
         if (active) {
-            UCT_CUDADRV_FUNC_LOG_ERR(
-                    cuIpcCloseMemHandle((CUdeviceptr)region->mapped_addr));
+            uct_cuda_ipc_close_memhandle(region);
         }
         ucs_free(region);
     }
     ucs_trace("%s: cuda ipc cache purged", cache->name);
 }
 
-static ucs_status_t uct_cuda_ipc_open_memhandle(const uct_cuda_ipc_rkey_t *key,
-                                                CUdeviceptr *mapped_addr)
+static ucs_status_t
+uct_cuda_ipc_open_memhandle_legacy(CUipcMemHandle memh,
+                                   CUdeviceptr *mapped_addr)
 {
     CUresult cuerr;
-    ucs_status_t status;
 
-    cuerr = cuIpcOpenMemHandle(mapped_addr, key->ph,
+    cuerr = cuIpcOpenMemHandle(mapped_addr, memh,
                                CU_IPC_MEM_LAZY_ENABLE_PEER_ACCESS);
-    if (cuerr == CUDA_SUCCESS) {
-        status = UCS_OK;
-    } else {
+    if (cuerr != CUDA_SUCCESS) {
         ucs_debug("cuIpcOpenMemHandle() failed: %s",
                   uct_cuda_base_cu_get_error_string(cuerr));
-        status = (cuerr == CUDA_ERROR_ALREADY_MAPPED) ? UCS_ERR_ALREADY_EXISTS :
-                                                        UCS_ERR_INVALID_PARAM;
+        return (cuerr == CUDA_ERROR_ALREADY_MAPPED) ?
+            UCS_ERR_ALREADY_EXISTS : UCS_ERR_INVALID_PARAM;
     }
 
+    return UCS_OK;
+}
+
+#if HAVE_CUDA_FABRIC
+static ucs_status_t uct_cuda_ipc_init_access_desc(CUmemAccessDesc *access_desc)
+{
+    access_desc->location.type = CU_MEM_LOCATION_TYPE_DEVICE;
+    access_desc->flags         = CU_MEM_ACCESS_FLAGS_PROT_READWRITE;
+
+    return UCT_CUDADRV_FUNC_LOG_ERR(cuCtxGetDevice(&access_desc->location.id));
+}
+
+static ucs_status_t
+uct_cuda_ipc_open_memhandle_vmm(uct_cuda_ipc_rkey_t *key,
+                                CUdeviceptr *mapped_addr)
+{
+    CUmemAccessDesc access_desc = {};
+    ucs_status_t status;
+    CUdeviceptr dptr;
+    CUmemGenericAllocationHandle handle;
+
+    status = UCT_CUDADRV_FUNC_LOG_ERR(cuMemImportFromShareableHandle(&handle,
+                (void*)&key->ph.handle.fabric_handle,
+                CU_MEM_HANDLE_TYPE_FABRIC));
+    if (status != UCS_OK) {
+        goto out;
+    }
+
+    status =
+        UCT_CUDADRV_FUNC_LOG_ERR(cuMemAddressReserve(&dptr, key->b_len, 0, 0, 0));
+    if (status != UCS_OK) {
+        goto release_handle;
+    }
+
+    status = UCT_CUDADRV_FUNC_LOG_ERR(cuMemMap(dptr, key->b_len, 0, handle, 0));
+    if (status != UCS_OK) {
+        goto release_va_range;
+    }
+
+    status = uct_cuda_ipc_init_access_desc(&access_desc);
+    if (status != UCS_OK) {
+        goto unmap_range;
+    }
+
+    status = UCT_CUDADRV_FUNC_LOG_ERR(cuMemSetAccess(dptr, key->b_len, &access_desc, 1));
+    if (status != UCS_OK) {
+        goto unmap_range;
+    }
+
+    *mapped_addr = dptr;
+
+    status = UCT_CUDADRV_FUNC_LOG_ERR(cuMemRelease(handle));
+    if (status == UCS_OK) {
+        goto out;
+    }
+
+unmap_range:
+    cuMemUnmap(dptr, key->b_len);
+release_va_range:
+    cuMemAddressFree(dptr, key->b_len);
+release_handle:
+    cuMemRelease(handle);
+out:
     return status;
 }
 
+static ucs_status_t cuda_ipc_rem_mpool_cache_create(uct_cuda_ipc_rkey_t *key,
+                                                    CUmemoryPool *mpool,
+                                                    CUdeviceptr *mapped_addr)
+{
+    CUmemAccessDesc access_desc = {};
+    CUdeviceptr dptr;
+    ucs_status_t status;
+
+    status = UCT_CUDADRV_FUNC_LOG_ERR(cuMemPoolImportFromShareableHandle(
+                mpool, (void *)&key->ph.handle.fabric_handle,
+                CU_MEM_HANDLE_TYPE_FABRIC, 0));
+    if (status != UCS_OK) {
+        goto err;
+    }
+
+    status = UCT_CUDADRV_FUNC_LOG_ERR(cuMemPoolImportPointer(&dptr, *mpool,
+                (CUmemPoolPtrExportData*)&key->ph.ptr));
+    if (status != UCS_OK) {
+        goto err_free_mpool;
+    }
+
+    status = uct_cuda_ipc_init_access_desc(&access_desc);
+    if (status != UCS_OK) {
+        goto err_free_ptr;
+    }
+
+    status = UCT_CUDADRV_FUNC_LOG_ERR(
+                cuMemPoolSetAccess(*mpool, &access_desc, 1));
+    if (status != UCS_OK) {
+        goto err_free_ptr;
+    }
+
+    *mapped_addr = dptr;
+    return UCS_OK;
+
+err_free_ptr:
+    cuMemFree(dptr);
+err_free_mpool:
+    cuMemPoolDestroy(*mpool);
+err:
+    return status;
+}
+
+static ucs_status_t
+uct_cuda_ipc_open_memhandle_mempool(uct_cuda_ipc_rkey_t *key,
+                                    CUdeviceptr *mapped_addr)
+{
+    khash_t(cuda_ipc_rem_mpool_cache) *hash = &uct_cuda_ipc_rem_mpool_cache.hash;
+    const CUmemFabricHandle *hkey           = &key->ph.handle.fabric_handle;
+    ucs_status_t status                     = UCS_OK;
+    CUmemoryPool mpool;
+    khiter_t khiter;
+    int khret;
+
+    pthread_rwlock_rdlock(&uct_cuda_ipc_rem_mpool_cache.lock);
+    khiter = kh_get(cuda_ipc_rem_mpool_cache, hash, *hkey);
+    if (ucs_likely(khiter != kh_end(hash))) {
+        key->ph.pool = kh_val(hash, khiter);
+        ucs_assert(key->ph.pool != NULL);
+        goto out_import_pointer;
+    }
+
+    /* acquire writer lock as we did not find cached mpool in hashmap */
+    pthread_rwlock_unlock(&uct_cuda_ipc_rem_mpool_cache.lock);
+    pthread_rwlock_wrlock(&uct_cuda_ipc_rem_mpool_cache.lock);
+
+    khiter = kh_put(cuda_ipc_rem_mpool_cache, hash, *hkey, &khret);
+    if ((khret == UCS_KH_PUT_BUCKET_EMPTY) ||
+        (khret == UCS_KH_PUT_BUCKET_CLEAR)) {
+        status = cuda_ipc_rem_mpool_cache_create(key, &mpool, mapped_addr);
+        if (status != UCS_OK) {
+            /* Remove stale key from the hash if failed to construct new value.
+             * Otherwise next lookup will return NULL value. */
+            kh_del(cuda_ipc_rem_mpool_cache, hash, khiter);
+            goto err;
+        }
+
+        kh_val(hash, khiter) = mpool;
+        key->ph.pool         = mpool;
+    } else if (khret == UCS_KH_PUT_KEY_PRESENT) {
+        key->ph.pool = kh_val(hash, khiter);
+    } else {
+        ucs_error("unable to use cuda_ipc remote_cache hash: %d", khret);
+        status = UCS_ERR_NO_RESOURCE;
+        goto err;
+    }
+
+out_import_pointer:
+    status = UCT_CUDADRV_FUNC_LOG_ERR(cuMemPoolImportPointer(mapped_addr,
+                key->ph.pool, (CUmemPoolPtrExportData*)&key->ph.ptr));
+
+err:
+    pthread_rwlock_unlock(&uct_cuda_ipc_rem_mpool_cache.lock);
+    return status;
+}
+#endif
+
+static ucs_status_t uct_cuda_ipc_open_memhandle(uct_cuda_ipc_rkey_t *key,
+                                                CUdeviceptr *mapped_addr)
+{
+
+    ucs_trace("key handle type %u", key->ph.handle_type);
+
+    switch(key->ph.handle_type) {
+    case UCT_CUDA_IPC_KEY_HANDLE_TYPE_LEGACY:
+        return uct_cuda_ipc_open_memhandle_legacy(key->ph.handle.legacy,
+                                                  mapped_addr);
+#if HAVE_CUDA_FABRIC
+    case UCT_CUDA_IPC_KEY_HANDLE_TYPE_VMM:
+        return uct_cuda_ipc_open_memhandle_vmm(key, mapped_addr);
+    case UCT_CUDA_IPC_KEY_HANDLE_TYPE_MEMPOOL:
+        return uct_cuda_ipc_open_memhandle_mempool(key, mapped_addr);
+#endif
+    default:
+        ucs_error("unsupported key handle type");
+        return UCS_ERR_INVALID_PARAM;
+    }
+}
+
 static void uct_cuda_ipc_cache_invalidate_regions(uct_cuda_ipc_cache_t *cache,
                                                   void *from, void *to)
 {
     ucs_list_link_t region_list;
     ucs_status_t status;
     uct_cuda_ipc_cache_region_t *region, *tmp;
+    int handle_type;
 
     ucs_list_head_init(&region_list);
     ucs_pgtable_search_range(&cache->pgtable, (ucs_pgt_addr_t)from,
@@ -134,8 +373,19 @@ static void uct_cuda_ipc_cache_invalidate_regions(uct_cuda_ipc_cache_t *cache,
             ucs_error("failed to remove address:%p from cache (%s)",
                       (void *)region->key.d_bptr, ucs_status_string(status));
         }
-        UCT_CUDADRV_FUNC_LOG_ERR(
-                cuIpcCloseMemHandle((CUdeviceptr)region->mapped_addr));
+
+        status = uct_cuda_ipc_close_memhandle(region);
+        if (status != UCS_OK) {
+#if HAVE_CUDA_FABRIC
+            handle_type = region->key.ph.handle_type;
+#else
+            handle_type = 1; /* legacy memory type is the only valid type */
+#endif
+            ucs_error("failed to close memhandle for base addr:%p type:%d (%s)",
+                      (void *)region->key.d_bptr, handle_type,
+                      ucs_status_string(status));
+        }
+
         ucs_free(region);
     }
     ucs_trace("%s: closed memhandles in the range [%p..%p]",
@@ -215,8 +465,7 @@ ucs_status_t uct_cuda_ipc_unmap_memhandle(pid_t pid, uintptr_t d_bptr,
                       (void *)region->key.d_bptr, ucs_status_string(status));
         }
         ucs_assert(region->mapped_addr == mapped_addr);
-        status = UCT_CUDADRV_FUNC_LOG_ERR(
-                cuIpcCloseMemHandle((CUdeviceptr)region->mapped_addr));
+        status = uct_cuda_ipc_close_memhandle(region);
         ucs_free(region);
     }
 
@@ -225,7 +474,7 @@ ucs_status_t uct_cuda_ipc_unmap_memhandle(pid_t pid, uintptr_t d_bptr,
 }
 
 UCS_PROFILE_FUNC(ucs_status_t, uct_cuda_ipc_map_memhandle, (key, mapped_addr),
-                 const uct_cuda_ipc_rkey_t *key, void **mapped_addr)
+                 uct_cuda_ipc_rkey_t *key, void **mapped_addr)
 {
     uct_cuda_ipc_cache_t *cache;
     ucs_status_t status;
@@ -243,8 +492,8 @@ UCS_PROFILE_FUNC(ucs_status_t, uct_cuda_ipc_map_memhandle, (key, mapped_addr),
                                   &cache->pgtable, key->d_bptr);
     if (ucs_likely(pgt_region != NULL)) {
         region = ucs_derived_of(pgt_region, uct_cuda_ipc_cache_region_t);
-        if (memcmp((const void *)&key->ph, (const void *)&region->key.ph,
-                   sizeof(key->ph)) == 0) {
+
+        if (key->ph.buffer_id == region->key.ph.buffer_id) {
             /*cache hit */
             ucs_trace("%s: cuda_ipc cache hit addr:%p size:%lu region:"
                       UCS_PGT_REGION_FMT, cache->name, (void *)key->d_bptr,
@@ -269,8 +518,7 @@ UCS_PROFILE_FUNC(ucs_status_t, uct_cuda_ipc_map_memhandle, (key, mapped_addr),
             }
 
             /* close memhandle */
-            UCT_CUDADRV_FUNC_LOG_ERR(
-                    cuIpcCloseMemHandle((CUdeviceptr)region->mapped_addr));
+            uct_cuda_ipc_close_memhandle(region);
             ucs_free(region);
         }
     }
@@ -412,11 +660,31 @@ void uct_cuda_ipc_destroy_cache(uct_cuda_ipc_cache_t *cache)
 UCS_STATIC_INIT {
     ucs_recursive_spinlock_init(&uct_cuda_ipc_remote_cache.lock, 0);
     kh_init_inplace(cuda_ipc_rem_cache, &uct_cuda_ipc_remote_cache.hash);
+
+#if HAVE_CUDA_FABRIC
+    pthread_rwlock_init(&uct_cuda_ipc_rem_mpool_cache.lock, NULL);
+    /* Assumption: If import process succeeds, then the two nodes are in the
+     * same domain. Within a domain, fabric handles are expected to be unique.
+     * For this reason, there is no need to maintain a hashmap per peer OS as
+     * key collisions are not expected to occur. */
+    kh_init_inplace(cuda_ipc_rem_mpool_cache, &uct_cuda_ipc_rem_mpool_cache.hash);
+#endif
 }
 
 UCS_STATIC_CLEANUP {
     uct_cuda_ipc_cache_t *rem_cache;
 
+#if HAVE_CUDA_FABRIC
+    CUmemoryPool mpool;
+
+    kh_foreach_value(&uct_cuda_ipc_rem_mpool_cache.hash, mpool, {
+        cuMemPoolDestroy(mpool);
+    });
+    kh_destroy_inplace(cuda_ipc_rem_mpool_cache,
+                       &uct_cuda_ipc_rem_mpool_cache.hash);
+    pthread_rwlock_destroy(&uct_cuda_ipc_rem_mpool_cache.lock);
+#endif
+
     kh_foreach_value(&uct_cuda_ipc_remote_cache.hash, rem_cache, {
         uct_cuda_ipc_destroy_cache(rem_cache);
     })
