@@ -18,12 +18,17 @@
 #include <ucs/async/async.h>
 #include <ucs/sys/compiler.h>
 #include <ucs/sys/string.h>
+#include <ucs/sys/netlink.h>
 #include <ucs/sys/sock.h>
 #include <ucs/sys/sys.h>
 #include <sys/poll.h>
 #include <libgen.h>
 #include <sched.h>
 
+#ifdef HAVE_NETLINK_RDMA
+#include <rdma/rdma_netlink.h>
+#endif
+
 
 /* This table is according to "Encoding for RNR NAK Timer Field"
  * in IBTA specification */
@@ -72,6 +77,14 @@ uct_ib_async_event_hash_equal(uct_ib_async_event_t event1,
 KHASH_IMPL(uct_ib_async_event, uct_ib_async_event_t, uct_ib_async_event_val_t, 1,
            uct_ib_async_event_hash_func, uct_ib_async_event_hash_equal)
 
+typedef struct uct_ib_device_subnet {
+    struct sockaddr_storage address;
+    unsigned                prefix_length;
+} uct_ib_device_subnet_t;
+
+UCS_ARRAY_DECLARE_TYPE(uct_ib_device_subnet_array_t, unsigned,
+                       uct_ib_device_subnet_t);
+
 #ifdef ENABLE_STATS
 static ucs_stats_class_t uct_ib_device_stats_class = {
     .name          = "",
@@ -136,6 +149,9 @@ static uct_ib_device_spec_t uct_ib_builtin_device_specs[] = {
   {"ConnectX-7", {0x15b3, 4129},
    UCT_IB_DEVICE_FLAG_MELLANOX | UCT_IB_DEVICE_FLAG_MLX5_PRM |
    UCT_IB_DEVICE_FLAG_DC_V2, 70},
+  {"ConnectX-8", {0x15b3, 4131},
+   UCT_IB_DEVICE_FLAG_MELLANOX | UCT_IB_DEVICE_FLAG_MLX5_PRM |
+   UCT_IB_DEVICE_FLAG_DC_V2, 80},
   {"BlueField", {0x15b3, 0xa2d2},
    UCT_IB_DEVICE_FLAG_MELLANOX | UCT_IB_DEVICE_FLAG_MLX5_PRM |
    UCT_IB_DEVICE_FLAG_DC_V2, 41},
@@ -875,9 +891,141 @@ int uct_ib_device_test_roce_gid_index(uct_ib_device_t *dev, uint8_t port_num,
     return 1;
 }
 
-ucs_status_t uct_ib_device_select_gid(uct_ib_device_t *dev, uint8_t port_num,
-                                      uct_ib_device_gid_info_t *gid_info)
+ucs_status_t
+uct_ib_device_roce_gid_to_sockaddr(sa_family_t af, const void *gid,
+                                   struct sockaddr_storage *sock_storage)
 {
+    struct sockaddr *sa = (struct sockaddr*)sock_storage;
+    const uint8_t *inet_addr;
+    size_t addr_size;
+    ucs_status_t status;
+
+    /* Set address family */
+    sa->sa_family = af;
+
+    /* Set port to 0 as it's not relevant for RoCE */
+    status = ucs_sockaddr_set_port(sa, 0);
+    if (status != UCS_OK) {
+        return status;
+    }
+
+    /* Get address size */
+    status = ucs_sockaddr_inet_addr_size(af, &addr_size);
+    if (status != UCS_OK) {
+        return status;
+    }
+
+    /* Set IP address */
+    inet_addr = UCS_PTR_BYTE_OFFSET(gid, sizeof(union ibv_gid) - addr_size);
+    return ucs_sockaddr_set_inet_addr(sa, inet_addr);
+}
+
+static ucs_status_t
+uct_ib_device_parse_subnet_filter(const ucs_config_allow_list_t *subnet_strs,
+                                  uct_ib_device_subnet_array_t *subnets)
+{
+    char *address_str, *mask_str;
+    char **subnet_str;
+    char subnet_str_dup[UCS_SOCKADDR_STRING_LEN];
+    uct_ib_device_subnet_t *subnet;
+    ucs_status_t status;
+
+    if (subnet_strs->mode == UCS_CONFIG_ALLOW_LIST_ALLOW_ALL) {
+        return UCS_OK;
+    }
+
+    ucs_carray_for_each(subnet_str, subnet_strs->array.names,
+                        subnet_strs->array.count) {
+        ucs_strncpy_safe(subnet_str_dup, *subnet_str, sizeof(subnet_str_dup));
+
+        /* Expect a string of the following pattern: x.x.x.x/y */
+        ucs_string_split(subnet_str_dup, "/", 2, &address_str, &mask_str);
+        if (mask_str == NULL) {
+            status = UCS_ERR_INVALID_PARAM;
+            goto err;
+        }
+
+        subnet = ucs_array_append_fixed(subnets);
+
+        /* Parse subnet address */
+        status = ucs_sock_ipstr_to_sockaddr(address_str, &subnet->address);
+        if (status != UCS_OK) {
+            goto err;
+        }
+
+        /* Parse subnet mask */
+        if (sscanf(mask_str, "%u", &subnet->prefix_length) != 1) {
+            status = UCS_ERR_INVALID_PARAM;
+            goto err;
+        }
+    }
+
+    return UCS_OK;
+
+err:
+    ucs_error("failed to parse RoCE subnet: %s", *subnet_str);
+    return status;
+}
+
+static int
+uct_ib_device_match_roce_subnet(const uct_ib_device_gid_info_t *gid_info,
+                                const uct_ib_device_subnet_array_t *subnets,
+                                ucs_config_allow_list_mode_t mode)
+{
+    const int is_allow_mode = (mode == UCS_CONFIG_ALLOW_LIST_ALLOW);
+    static const char UCS_V_UNUSED *allow_mode_str[] = {"accepted",
+                                                        "restricted"};
+    const uct_ib_device_subnet_t *subnet;
+    struct sockaddr_storage gid_sockaddr;
+    char gid_str[UCS_SOCKADDR_STRING_LEN];
+    char subnet_str[UCS_SOCKADDR_STRING_LEN];
+
+    if (mode == UCS_CONFIG_ALLOW_LIST_ALLOW_ALL) {
+        return 1;
+    }
+
+    /* Convert GID to sockaddr structure */
+    if (uct_ib_device_roce_gid_to_sockaddr(gid_info->roce_info.addr_family,
+                                           &gid_info->gid,
+                                           &gid_sockaddr) != UCS_OK) {
+        ucs_error("failed to convert GID %u to sockaddr", gid_info->gid_index);
+        return 0;
+    }
+
+    /* Iterate over all subnets and compare them with GID */
+    ucs_array_for_each(subnet, subnets) {
+        if (!ucs_sockaddr_is_same_subnet(
+                    (const struct sockaddr*)&gid_sockaddr,
+                    (const struct sockaddr*)&subnet->address,
+                    subnet->prefix_length)) {
+            continue;
+        }
+
+        ucs_sockaddr_str((const struct sockaddr*)&gid_sockaddr, gid_str,
+                         UCS_SOCKADDR_STRING_LEN);
+        ucs_sockaddr_str((const struct sockaddr*)&subnet->address, subnet_str,
+                         UCS_SOCKADDR_STRING_LEN);
+        ucs_trace("address %s at gid[%u] was %s by subnet filter %s/%u",
+                  gid_str, gid_info->gid_index, allow_mode_str[!is_allow_mode],
+                  subnet_str, subnet->prefix_length);
+
+        /* Accept/Restrict GID according to required policy */
+        return is_allow_mode;
+    }
+
+    ucs_trace("gid index %u was %s due to no matching subnets",
+              gid_info->gid_index, allow_mode_str[!is_allow_mode]);
+
+    /* Handle non-matched GID according to policy */
+    return !is_allow_mode;
+}
+
+ucs_status_t
+uct_ib_device_select_gid(uct_ib_device_t *dev, uint8_t port_num,
+                         const ucs_config_allow_list_t *subnet_strs,
+                         uct_ib_device_gid_info_t *gid_info)
+{
+    static const size_t max_str_len                     = 200;
     static const uct_ib_roce_version_info_t roce_prio[] = {
         {UCT_IB_DEVICE_ROCE_V2, AF_INET},
         {UCT_IB_DEVICE_ROCE_V2, AF_INET6},
@@ -887,11 +1035,19 @@ ucs_status_t uct_ib_device_select_gid(uct_ib_device_t *dev, uint8_t port_num,
     int gid_tbl_len         = uct_ib_device_port_attr(dev, port_num)->gid_tbl_len;
     ucs_status_t status     = UCS_OK;
     int priorities_arr_len  = ucs_static_array_size(roce_prio);
+    UCS_ARRAY_DEFINE_ONSTACK(uct_ib_device_subnet_array_t, subnets,
+                             subnet_strs->array.count);
     uct_ib_device_gid_info_t gid_info_tmp;
-    int i, prio_idx;
+    int i, prio_idx, res;
+    char subnet_list_str[max_str_len];
 
     ucs_assert(uct_ib_device_is_port_roce(dev, port_num));
 
+    status = uct_ib_device_parse_subnet_filter(subnet_strs, &subnets);
+    if (status != UCS_OK) {
+        return status;
+    }
+
     /* search for matching GID table entries, according to the order defined
      * in priorities array
      */
@@ -906,8 +1062,9 @@ ucs_status_t uct_ib_device_select_gid(uct_ib_device_t *dev, uint8_t port_num,
 
             if ((roce_prio[prio_idx].ver         == gid_info_tmp.roce_info.ver) &&
                 (roce_prio[prio_idx].addr_family == gid_info_tmp.roce_info.addr_family) &&
-                uct_ib_device_test_roce_gid_index(dev, port_num, &gid_info_tmp.gid, i)) {
-
+                uct_ib_device_test_roce_gid_index(dev, port_num, &gid_info_tmp.gid, i) &&
+                uct_ib_device_match_roce_subnet(&gid_info_tmp, &subnets,
+                                                subnet_strs->mode)) {
                 gid_info->gid_index = i;
                 gid_info->roce_info = gid_info_tmp.roce_info;
                 goto out_print;
@@ -915,6 +1072,15 @@ ucs_status_t uct_ib_device_select_gid(uct_ib_device_t *dev, uint8_t port_num,
         }
     }
 
+    if (subnet_strs->mode != UCS_CONFIG_ALLOW_LIST_ALLOW_ALL) {
+        res = ucs_config_sprintf_allow_list(subnet_list_str, max_str_len,
+                                            subnet_strs,
+                                            &ucs_config_array_string);
+        ucs_error("failed to find a gid which matches/unmatches the following "
+                  "subnet list: %s", res ? subnet_list_str : "<none>");
+        return UCS_ERR_INVALID_PARAM;
+    }
+
     gid_info->gid_index             = UCT_IB_DEVICE_DEFAULT_GID_INDEX;
     gid_info->roce_info.ver         = UCT_IB_DEVICE_ROCE_V1;
     gid_info->roce_info.addr_family = AF_INET;
@@ -1125,9 +1291,9 @@ ucs_status_t uct_ib_device_mtu(const char *dev_name, uct_md_h md, int *p_mtu)
     return UCS_OK;
 }
 
-int uct_ib_device_is_gid_raw_empty(uint8_t *gid_raw)
+int uct_ib_device_is_gid_valid(const union ibv_gid *gid)
 {
-    return (*(uint64_t *)gid_raw == 0) && (*(uint64_t *)(gid_raw + 8) == 0);
+    return gid->global.interface_id != 0;
 }
 
 ucs_status_t uct_ib_device_query_gid(uct_ib_device_t *dev, uint8_t port_num,
@@ -1143,7 +1309,7 @@ ucs_status_t uct_ib_device_query_gid(uct_ib_device_t *dev, uint8_t port_num,
         return status;
     }
 
-    if (uct_ib_device_is_gid_raw_empty(gid_info.gid.raw)) {
+    if (!uct_ib_device_is_gid_valid(&gid_info.gid)) {
         ucs_log(error_level, "invalid gid[%d] on %s:%d", gid_index,
                 uct_ib_device_name(dev), port_num);
         return UCS_ERR_INVALID_ADDR;
@@ -1343,3 +1509,62 @@ const char* uct_ib_ah_attr_str(char *buf, size_t max,
 
     return buf;
 }
+
+#ifdef HAVE_NETLINK_RDMA
+static ucs_status_t
+uct_ib_device_is_smi_cb(const struct nlmsghdr *nlh, void *arg)
+{
+    int *is_smi_p = (int*)arg;
+    const struct nlattr *attr;
+    uint8_t dev_type;
+
+    for (attr = NLMSG_DATA(nlh); UCS_PTR_BYTE_DIFF(nlh, attr) < nlh->nlmsg_len;
+         attr = UCS_PTR_BYTE_OFFSET(attr, NLA_ALIGN(attr->nla_len))) {
+        if (attr->nla_type == RDMA_NLDEV_ATTR_DEV_TYPE /* 99 */) {
+            dev_type = *(const uint8_t*)UCS_PTR_BYTE_OFFSET(attr, NLA_HDRLEN);
+            if (dev_type == RDMA_DEVICE_TYPE_SMI /* 1 */) {
+                *is_smi_p = 1;
+                return UCS_OK;
+            }
+        }
+    }
+
+    return UCS_INPROGRESS;
+}
+
+int uct_ib_device_is_smi(struct ibv_device *ibv_device)
+{
+    struct nlattr *attr;
+    uint32_t *dev_index_attr;
+    size_t header_length;
+    ucs_status_t status;
+    int is_smi;
+
+    header_length   = NLA_HDRLEN + sizeof(*dev_index_attr);
+    attr            = ucs_alloca(header_length);
+    dev_index_attr  = (uint32_t*)UCS_PTR_BYTE_OFFSET(attr, NLA_HDRLEN);
+    attr->nla_type  = RDMA_NLDEV_ATTR_DEV_INDEX;
+    attr->nla_len   = header_length;
+    *dev_index_attr = ibv_get_device_index(ibv_device);
+    if (*dev_index_attr == -1) {
+        ucs_debug("%s: failed to get device index",
+                  ibv_get_device_name(ibv_device));
+        return 0;
+    }
+
+    is_smi = 0;
+    status = ucs_netlink_send_request(
+            NETLINK_RDMA, RDMA_NL_GET_TYPE(RDMA_NL_NLDEV, RDMA_NLDEV_CMD_GET),
+            0, attr, header_length, uct_ib_device_is_smi_cb, &is_smi);
+    if (status != UCS_OK) {
+        return 0;
+    }
+
+    return is_smi;
+}
+#else
+int uct_ib_device_is_smi(struct ibv_device *ibv_device)
+{
+    return 0;
+}
+#endif
