@@ -118,12 +118,15 @@ uct_rc_verbs_ep_atomic_post(uct_rc_verbs_ep_t *ep, int opcode, uint64_t compare_
                             uint64_t swap, uint64_t remote_addr, uct_rkey_t rkey,
                             uct_rc_iface_send_desc_t *desc, int force_sig)
 {
+    uint32_t ib_rkey = uct_ib_resolve_atomic_rkey(rkey,
+                                                  ep->super.atomic_mr_offset,
+                                                  &remote_addr);
     struct ibv_send_wr wr;
     struct ibv_sge sge;
 
+
     UCT_RC_VERBS_FILL_ATOMIC_WR(wr, wr.opcode, sge, (enum ibv_wr_opcode)opcode,
-                                compare_add, swap, remote_addr,
-                                uct_ib_md_direct_rkey(rkey));
+                                compare_add, swap, remote_addr, ib_rkey);
     UCT_TL_EP_STAT_ATOMIC(&ep->super.super);
     uct_rc_verbs_ep_post_send_desc(ep, &wr, desc, force_sig, INT_MAX);
     uct_rc_ep_enable_flush_remote(&ep->super);
@@ -616,7 +619,8 @@ int uct_rc_verbs_ep_is_connected(const uct_ep_h tl_ep,
         addr_qp = uct_ib_unpack_uint24(rc_addr->qp_num);
     }
 
-    return uct_rc_ep_is_connected(&ah_attr, params, qp_num, addr_qp);
+    return uct_rc_ep_is_connected(&ep->super, &ah_attr, params, qp_num,
+                                  addr_qp);
 }
 
 ucs_status_t
